from locale import strcoll
from logging import log
from wordcloud import WordCloud
import matplotlib.pyplot as plt
from nltk.tokenize import RegexpTokenizer
from nltk import bigrams, trigrams
from nltk.probability import FreqDist
import pandas as pd
import numpy as np
import seaborn as sns
from pyspark.sql import DataFrame, dataframe, functions as F, types as T
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.manifold import TSNE
from sklearn.decomposition import PCA
from gensim.models import word2vec
import multiprocessing
from time import time
import nltk
from api_model.utils.logger import logger
from sklearn.cluster import KMeans
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.feature_extraction.text import TfidfVectorizer
from kneebow.rotor import Rotor
from sklearn.cluster import MiniBatchKMeans
from api_model.utils.functions import TransforDatas


nltk.download('punkt')

NLTK_STOPWORDS = nltk.corpus.stopwords.words('portuguese')

PATH_READ = '/content/drive/My Drive/'
#PATH_READ = '/opt/dna/find-keywords/datalake'

class NlVisualization:


    def __init__(self, 
                filename: str,
                column_filter: str,
                column_text: str,
                whats_process: str):
        
        self.filename = filename
        self.column_filter = column_filter
        self.column_text = column_text
        self.whats_process = whats_process

    @classmethod
    def wordCloud_Topics(self,
                filename: str,
                column_filter: str,              
                column_text: str,
                whats_process: str):
    

        df = pd.read_csv(f"{PATH_READ}/{filename}_tratado.csv", sep=';', encoding='utf-8')

        if column_filter == '':
            pass
        else:
            df = df[df[column_filter].notnull()]
            df = df[df[column_filter].notna()]
            df = df[df[column_filter] != '']
            df = df[df[column_filter] !='[]']

        list_str = df[column_text].values.tolist()
        sent = ','.join([str(i) for i in list_str])

        tokenizer = RegexpTokenizer(r'\w+')
        sent_words = tokenizer.tokenize(sent)

        if whats_process == 'bigram':
            brigram = bigrams(sent_words)
            join_bigram = [' '.join((a, b)) for a, b in brigram]
        else:
            brigram = trigrams(sent_words)
            join_bigram = [' '.join((a, b, c)) for a, b, c in brigram]

        freq_dist = FreqDist(join_bigram)

        wc = WordCloud(
            min_font_size=14,
            max_font_size=170,
            max_words=50,
            background_color='white',
            colormap='winter',
            width=700,
            height=550,
            stopwords=NLTK_STOPWORDS,
        )

        wordcloud_bigrama = wc.generate_from_frequencies(freq_dist)
        plt.figure(2, figsize=(19, 19 / 1.6180))
        plt.imshow(wordcloud_bigrama, interpolation='bilinear')
        plt.tight_layout(pad=0)
        plt.axis('off')

        plt.show()        


        return None
    
    @classmethod
    def plot_10_most_common_words(self,
                filename: str,
                column_filter: str,              
                column_text: str):
        
        df = pd.read_csv(f"{PATH_READ}/{filename}_tratado.csv", sep=';', encoding='utf-8')

        if column_filter == '':
            pass
            df = df[~df[column_text].isnull()]
        else:                
            df = df[df[column_filter].notnull()]
            df = df[df[column_filter].notna()]
            df = df[df[column_filter] != '']

        number_cluster = 20
        list_str = df[column_text].values.tolist()

        vectorizer = CountVectorizer(stop_words=NLTK_STOPWORDS)
        count_vect = vectorizer.fit_transform(list_str)

        words = vectorizer.get_feature_names()
        total_counts = np.zeros(len(words))

        for t in count_vect:
            total_counts += t.toarray()[0]

        count_dict = zip(words, total_counts)
        count_dict = sorted(count_dict, key=lambda x: x[1], reverse=True)[0:number_cluster]
        words = [w[0] for w in count_dict]
        counts = [w[1] for w in count_dict]
        x_pos = np.arange(len(words))

        plt.figure(2, figsize=(19, 19 / 1.6180))
        plt.subplot(title=str(number_cluster) + ' most common words')

        sns.set_context('notebook', font_scale=1.75, rc={'lines.linewidth': 2.5})
        sns.barplot(x_pos, counts, palette='husl')
        plt.xticks(x_pos, words, rotation=90)
        plt.xlabel('words')
        plt.ylabel('counts')

        plt.show()

        return None

    @classmethod
    def pareto_plot(self,
        filename: str,
        x: str = None,
        y: str = None,
        title: str = None,
        limite: int = None,
        show_pct_y: bool = False,
        pct_format: str = '{0:.0%}',
    ):

        df = pd.read_csv(f"{PATH_READ}/{filename}_tratado.csv", sep=';', encoding='utf-8')
        
        logger.info('plot graph of pareto with categorized')
        df = df[x].value_counts().head(20)
        tmp = pd.DataFrame({x: df.index, 'count':df.values})
        
        xlabel = x
        ylabel = y
        x = tmp[x].values
        y = tmp[y].values
        weights = y / y.sum()
        cumsum = weights.cumsum()

        fig, ax1 = plt.subplots()
        ax1.bar(x, y)
        ax1.set_xlabel(xlabel)
        ax1.set_ylabel(ylabel)
        ax1.tick_params('x', rotation=45)

        ax2 = ax1.twinx()
        ax2.plot(x, cumsum, '-ro', alpha=0.5)
        ax2.set_ylabel('', color='r')
        ax2.tick_params('y', colors='r')

        plt.rcParams['figure.figsize'] = (46, 9)
        vals = ax2.get_yticks()
        ax2.set_yticklabels(['{:,.2%}'.format(x) for x in vals])

        # hide y-labels on right side
        if not show_pct_y:
            ax2.set_yticks([])

        formatted_weights = [pct_format.format(x) for x in cumsum]
        for i, txt in enumerate(formatted_weights):
            ax2.annotate(txt, (x[i], cumsum[i]), fontweight='heavy')

        if title:
            plt.title(title)

        plt.show()

    @classmethod
    def model_2vectors(self,
        filename: str, content_column: str = None, vector_size: int = None, min_count: int = None, window: int = None
    ):
        cores = multiprocessing.cpu_count()  # Count the number of cores in a computer

        model = word2vec.Word2Vec(
            min_count=min_count,
            window=window,
            vector_size=vector_size,
            sample=6e-5,
            alpha=0.03,
            min_alpha=0.0007,
            negative=20,
            workers=cores - 1,
        )

        return model

    @classmethod
    def tsne_plot(self,
        filename:str,
        content_column: str,
        word: str,
        k: int = None,
        n_iter: int = None,
        vector_size: int = None,
        min_count: int = None,
        window: int = None,
    ):

        logger.info('plot graph of maps words')

        df = pd.read_csv(f"{PATH_READ}/{filename}_tratado.csv", sep=';', encoding='utf-8')
        # tokenizer column to model word2vec
        df.loc[:, f'{content_column}_len'] = df.apply(lambda row: nltk.word_tokenize(row[content_column]), axis=1)

        model = self.model_2vectors(df, None, vector_size=vector_size, min_count=min_count, window=window)

        t = time()
        model.build_vocab(df[f'{content_column}_len'], progress_per=10000)
        logger.debug(f'Time to build vocab: {format(round((time() - t) / 60, 2))}')

        list_names = [t[0] for t in model.wv.most_similar(positive=[word], topn=k)][10:]

        arrays = np.empty((0, vector_size), dtype='f')
        word_labels = [word]
        color_list = ['red']

        # adds the vector of the query word
        arrays = np.append(arrays, model.wv.__getitem__([word]), axis=0)
        # gets list of most similar words
        close_words = model.wv.most_similar([word])

        # adds the vector for each of the closest words to the array
        for wrd_score in close_words:
            wrd_vector = model.wv.__getitem__([wrd_score[0]])
            word_labels.append(wrd_score[0])
            color_list.append('blue')
            arrays = np.append(arrays, wrd_vector, axis=0)

        # adds the vector for each of the words from list_names to the array
        for wrd in list_names:
            wrd_vector = model.wv.__getitem__([wrd])
            word_labels.append(wrd)
            color_list.append('green')
            arrays = np.append(arrays, wrd_vector, axis=0)

        # Reduces the dimensionality from 300 to 50 dimensions with PCA
        reduc = PCA(n_components=20, svd_solver='full').fit_transform(arrays)
        # Finds t-SNE coordinates for 2 dimensions
        np.set_printoptions(suppress=True)

        tsne_model = TSNE(perplexity=40, n_components=2, init='pca', n_iter=n_iter).fit_transform(reduc)

        # Sets everything up to plot
        df = pd.DataFrame(
            {
                'x': [x for x in tsne_model[:, 0]],
                'y': [y for y in tsne_model[:, 1]],
                'words': word_labels,
                'color': color_list,
            }
        )

        fig, _ = plt.subplots()
        fig.set_size_inches(9, 9)

        # Basic plot
        p1 = sns.regplot(data=df, x='x', y='y', fit_reg=False, marker='o', scatter_kws={'s': 40, 'facecolors': df['color']})

        # Adds annotations one by one with a loop
        for line in range(0, df.shape[0]):
            p1.text(
                df['x'][line],
                df['y'][line],
                '  ' + df['words'][line].title(),
                horizontalalignment='left',
                verticalalignment='bottom',
                size='medium',
                color=df['color'][line],
                weight='normal',
            ).set_size(15)

        plt.xlim(tsne_model[:, 0].min() - 50, tsne_model[:, 0].max() + 50)
        plt.ylim(tsne_model[:, 1].min() - 50, tsne_model[:, 1].max() + 50)

        plt.title('t-SNE visualization for {}'.format(word.title()))

    @classmethod
    def tsne_plot_all(self,
        filename:str,
        content_column: str,
        n_iter: int = None,
        vector_size: int = None,
        min_count: int = None,
        window: int = None,
    ):

        logger.info('plot graph of maps all of words')

        df = pd.read_csv(f"{PATH_READ}/{filename}_tratado.csv", sep=';', encoding='utf-8')
        # tokenizer column to model word2vec
        df.loc[:, f'{content_column}_len'] = df.apply(lambda row: nltk.word_tokenize(row[content_column]), axis=1)

        model = self.model_2vectors(df, f'{content_column}_len', vector_size=vector_size, min_count=min_count, window=window)

        labels = []
        tokens = []

        for word in model.wv.key_to_index:
            tokens.append(model.wv.key_to_index[word])
            labels.append(word)

        tsne_model = TSNE(perplexity=40, n_components=1, init='pca', n_iter=n_iter)
        tk = np.array(tokens).reshape(1, -1)
        new_values = tsne_model.fit_transform(tk)
        x = []
        y = []

        for value in new_values:
            x.append(value[0])
            y.append(value[1])

        plt.figure(2, figsize=(38, 28 / 1.6180))
        for i in range(len(x)):
            plt.scatter(x[i], y[i])
            plt.annotate(labels[i], xy=(x[i], y[i]), xytext=(18, 12), textcoords='offset points', ha='right', va='bottom')

        plt.show()

    @classmethod
    def get_freq_table(filename:str, coutent_column: str):

        logger.info('plot table with frequency of labels')

        df = pd.read_csv(f"{PATH_READ}/{filename}_tratado.csv", sep=';', encoding='utf-8')

        counts = df[coutent_column].value_counts()
        freq = pd.DataFrame(
            {
                coutent_column: counts.index,
                'Frequência Absoluta': counts.values,
                'Frequência Relativa': ['{:.2f}'.format(100 * v) for v in counts.values / counts.sum()],
            }
        )
        freq = freq.set_index(freq[coutent_column])
        freq.drop(coutent_column, 1, inplace=True)

        return freq

    @classmethod
    def plot_elbow(self, K: int = None, sum_of_squared_distances: float = None, optimal_k: int = None):

        plt.plot(range(1, K + 1), sum_of_squared_distances, 'bx-')
        plt.xlabel('k')
        plt.ylabel('Sum_of_squared_distances')
        plt.title('Elbow Method For Optimal k')

        plt.axvline(x=optimal_k)
        plt.show()


    @classmethod
    def plot_kmeans_clustering(self, resume_features: np.array = [], kmens_label: np.array = []):

        plt.figure('Cluster K-Means')
        plt.scatter(resume_features[:, 0], resume_features[:, 1], c=kmens_label)
        plt.xlabel('Dividend Yield')
        plt.ylabel('Returns')
        plt.title('Cluster K-Means')
        plt.show()


    @classmethod
    def clustering_model(self,
        filename: strcoll, content_column: str = None, model: str = 'kmeans', plot: bool = True, max_k: int = None, **kwargs
    ):

        df = pd.read_csv(f"{PATH_READ}/{filename}_tratado.csv", sep=';', encoding='utf-8')
        # TO-DO : Add Elbow Method
        logger.info('Start optmal clustering of kmeans model')
        if model == 'kmeans':

            #data = df.select(content_column).rdd.flatMap(lambda x: x).collect()
            data = df[content_column].values

            tfidf = TfidfVectorizer(use_idf=True)
            features = tfidf.fit_transform(data)

            logger.info('Find number of clusters')
            sum_of_squared_distances = []
            K = max_k
            for k in range(1, K + 1):
                km = MiniBatchKMeans(n_clusters=k, init='k-means++', n_init=1, init_size=1000, batch_size=1000, **kwargs)
                km = km.fit(features)
                sum_of_squared_distances.append(km.inertia_)

            rotor = Rotor()

            data = [[-sum_of_squared_distances[x], x] for x in range(K)]
            rotor.fit_rotate(data)

            optimal_k = rotor.get_elbow_index() + 1

            if plot:
                self.plot_elbow(K, sum_of_squared_distances, optimal_k)

                logger.info(f'Optimal number of clusters: {optimal_k}')
                clst = MiniBatchKMeans(n_clusters=optimal_k, init='k-means++', n_init=1, init_size=1000, batch_size=1000)

            logger.info('Using model of optmal clustering finded')
            nc = range(1, optimal_k)
            kmeans = [KMeans(n_clusters=i).fit(features) for i in nc]

            logger.info('Dimension of Reduction from compreendition vectorial space')
            c = len(kmeans) - 1
            resumed_features = LinearDiscriminantAnalysis(n_components=2).fit_transform(
                features.toarray(), kmeans[c].labels_
            )
            resumed_features.shape

            logger.info('plot scatter of clustering')
            self.plot_kmeans_clustering(resumed_features, kmeans[c].labels_)

        else:
            logger.info('Please Select An Available Clustering Method.')
            return []

        clst.fit(data)

        logger.info(f'Predict model in dataframe, {kmeans[c].labels_}')

        # classifier_udf = F.udf(
        #     lambda text: str(kmeans[c].predict(tfidf.transform([text]))[0]) if text else None, T.StringType()
        # )

        # df = df.withColumn('ml_kmeans', classifier_udf(content_column))
        # logger.info('Finishing process')

        df['ml_kmeans'] = kmeans[c].labels_

        df = TransforDatas.save_file(df, filename=filename, meth=None)
        logger.info('New DataFrame Saved with Kmeans leabels')

        return clst, optimal_k, df       


from api_model.nlextract import NLExtractor
import pandas as pd
from api_model.utils.functions import TransforDatas
from pyspark.sql import SparkSession, functions as F, types as T
from api_model.utils.logger import logger, Colorize


class NlExtractorProcess(NLExtractor):

    def __init__(self, 
                filename: str,
                prefix: str,
                prefix_sep: str,
                column_text: str,
                whats_process: str,
                list_pattern: dict,
                id_database: str,
                type_find: str,
                additional_stop_words: list,
                activate_stopwords: str,
                interlocutor: dict,
                response_time: str,
                format_data: str,
                encoding: str
                ):
            
        self.filename = filename
        self.prefix = prefix
        self.prefix_sep = prefix_sep
        self.column_text = column_text
        self.whats_process = whats_process
        self.list_pattern = list_pattern
        self.id_database = id_database
        self.type_find = type_find
        self.additional_stop_words = additional_stop_words
        self.activate_stopwords = activate_stopwords
        self.interlocutor = interlocutor
        self.response_time = response_time
        self.format_data = format_data
        self.encoding = encoding

        super().__init__()
    
    @classmethod
    def call_process(self,
        filename,
        prefix,
        prefix_sep,
        column_text,
        whats_process,
        list_pattern,
        id_database,
        type_find,
        additional_stop_words,
        activate_stopwords,
        interlocutor,
        response_time,
        format_data,
        encoding):
        
        """
        whats_process = 'complete'
            return: process all pipeline
        whats_process = 'partial'
            return: findkeywords and process bigrams
        whats_process = 'only_keywords'
            return: findkeywords       
        """

        logger.info('Load CSV')
        df = TransforDatas.load_file(prefix, filename, prefix_sep, encoding, meth=None)

        logger.info('Normalize Datas Values')
        df = TransforDatas.normalize_data(df=df, column_text=column_text, response_time=response_time, format_data=format_data, id_database=id_database)
      
        if whats_process == 'complete':
            logger.info(f'Start Complete Process')

            if activate_stopwords == 'sim':
                logger.info('Using StopWords')
                df = TransforDatas.stop_words_text(df=df, column_text=column_text, additional_stop_words=additional_stop_words)

            logger.info('Start Word Search')
            df = TransforDatas.word_search(df=df, list_pattern=list_pattern, type_find=type_find, column_text=column_text)                

            logger.info('Start Text Mining')
            df = TransforDatas.text_mining(df=df, column_text=column_text)

            
            logger.info('Send Some Statistics of DataFrame')
            df = TransforDatas.statistics_dataframe(df=df, column_text=column_text)

            logger.info('Called Pyspark DataFrame')
            sparkDF = TransforDatas.convert_dataframe(df=df, id_database=id_database, column_text=column_text, response_time=response_time, filename=filename, prefix=prefix, prefix_sep=prefix_sep, interlocutor=interlocutor, encoding=encoding)

            logger.info('remove null values of dataset')
            sparkDF = sparkDF.filter((F.col('message_content') != '')
                        & (F.col('message_content') != ' ')
                        & (F.col('message_content').isNotNull()))

            logger.debug(f'count rows after remove null values {sparkDF.count()}')   

            logger.debug('created a new collect dict of interlocutor')
            out = dict()
            for key, index in interlocutor.items():
                out = {'message_author':index}
            logger.debug(f'new collect dict {out}')

            logger.debug('created a original column')
            if 'original_message' not in df.columns:
                sparkDF = sparkDF.withColumn('original_message', F.col('message_content'))            
            
            logger.info('agroup all menssages for ticket')
            sparkDF = self.group_df(df=sparkDF, interlocutor=out, message_content='message_content')       

            logger.info(f'numbers of rows agrouped {sparkDF.count()}')

            output_prefix =  'countent'
            logger.info(f'Generating wordcloud columns for "{Colorize.get_color("all_messages", color="cyan")}" with prefix "{output_prefix}"')           
            sparkDF, _ = self.most_relevant_ngram(
                sparkDF, 'all_messages', id_field=id_database, output_column_prefix=output_prefix
            )

            logger.debug(f'columns process in most relevant ngrams {sparkDF.printSchema()}')
            df = sparkDF.toPandas()            

        if whats_process == 'partial':
            print(f'Start Partial Process')

            if activate_stopwords == 'sim':
                logger.info('Using StopWords')
                df = TransforDatas.stop_words_text(df=df, column_text=column_text, additional_stop_words=additional_stop_words)

            logger.info('Start Word Search')
            df = TransforDatas.word_search(df=df, list_pattern=list_pattern, type_find=type_find, column_text=column_text)
            
            logger.info('Send Some Statistics of DataFrame')
            df = TransforDatas.statistics_dataframe(df=df, column_text=column_text)

            logger.info('Called Pyspark DataFrame')
            sparkDF = TransforDatas.convert_dataframe(df=df, id_database=id_database, column_text=column_text, response_time=response_time, filename=filename, prefix=prefix, prefix_sep=prefix_sep, interlocutor=interlocutor)

            logger.info('remove null values of dataset')
            sparkDF = sparkDF.filter((F.col('message_content') != '')
                        & (F.col('message_content') != ' ')
                        & (F.col('message_content').isNotNull()))

            logger.debug(f'count rows after remove null values {sparkDF.count()}')   

            logger.debug('created a new collect dict of interlocutor')
            out = dict()
            for key, index in interlocutor.items():
                out = {'message_author':index}
            logger.debug(f'new collect dict {out}')

            logger.debug('created a original column')
            if 'original_message' not in df.columns:
                sparkDF = sparkDF.withColumn('original_message', F.col('message_content'))            
            
            logger.info('agroup all menssages for ticket')
            sparkDF = self.group_df(df=sparkDF, interlocutor=out, message_content='message_content')       

            logger.info('process bigrams and trigrams of column_text')
            output_prefix =  'countent'
            sparkDF, _ = self.most_relevant_ngram(
                sparkDF, 'all_messages', id_field=id_database, output_column_prefix=output_prefix
            )
            df = sparkDF.toPandas()
            # logger.info('Merge DataFrames')
            # df = TransforDatas.merge_dataframe(df=df_pandas, df_all=df_all)

        if whats_process == 'only_keywords':
            print(f'Start Only KeyWords Find Process')

            if activate_stopwords == 'sim':
                logger.info('Using StopWords')
                df = TransforDatas.stop_words_text(df=df, column_text=column_text, additional_stop_words=additional_stop_words)

            logger.info('Start Word Search')
            df = TransforDatas.word_search(df=df, list_pattern=list_pattern, type_find=type_find, column_text=column_text)
            
            logger.info('Send Some Statistics of DataFrame')
            df = TransforDatas.statistics_dataframe(df=df, column_text=column_text)

        logger.info('Finishing Process and Save csv File')
        df = TransforDatas.save_file(df, filename=filename, meth=None)
        logger.info(f'schema of dataframe saved: {df.info()}')

        TransforDatas.delete_file(filename=f'{filename}_temp')
        logger.debug('file temp deleted')        

        return df


__all__ = ['call_process']


from api_model.utils.colorize import Colorize
from datetime import datetime
from pytz import timezone


sao_paulo_timezone = timezone('America/Sao_Paulo')


class Logger:
    """
    Logger class

    print only logs with criticity_level equal or less level informed
    """

    current_level = 5
    levels = ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET']

    def __init__(self, level='NOTSET'):
        self.set_level(level)

    def set_level(self, level):
        try:
            self.current_level = self.levels.index(level)
        except ValueError:
            self.error(f'{level} is not a accepted level. Try one of this: {self.levels}')

    @staticmethod
    def _execute(msg):
        print(f'[{datetime.now().astimezone(sao_paulo_timezone).strftime("%m/%d %H:%M:%S")}] - {msg}')

    def debug(self, msg):
        if self.levels.index('DEBUG') <= self.current_level:
            self._execute(f"{Colorize.get_color('[Debug]: ', color='green', style='bold')} {msg}")

    def info(self, msg):
        if self.levels.index('INFO') <= self.current_level:
            self._execute(f"{Colorize.get_color('[Info]: ', color='cyan', style='bold')} {msg}")

    def warn(self, msg):
        if self.levels.index('WARNING') <= self.current_level:
            self._execute(f"{Colorize.get_color('[Warning]: ', color='yellow', style='bold')} {msg}")

    def error(self, msg):
        if self.levels.index('ERROR') <= self.current_level:
            self._execute(f"{Colorize.get_color('[Error]: ', color='red', style='blink')} {msg}")

    def critical(self, msg):
        if self.levels.index('CRITICAL') <= self.current_level:
            self._execute(f"{Colorize.get_color('[CRITICAL]: ', color='on_red', style='blink')} {msg}")


logger = Logger(level='DEBUG')

__all__ = ['logger']



class Colorize:
    styles = {
        'default': 0,  # Turn off all attributes
        'bold': 1,  # Set bold mode
        'unable': 2,  # Set unable mode
        'italic': 3,  # Set italic mode
        'underline': 4,  # Set underline mode
        'blink': 5,  # Set blink mode
        'swap': 7,  # Exchange foreground and background colors
        # Hide text (foreground color would be the same as background)
        'hide': 8,
        'strykethrough': 9,  # Exchange foreground and background colors
    }

    colors = {
        'black': 30,
        'red': 31,
        'green': 32,
        'yellow': 33,
        'blue': 34,
        'magenta': 35,
        'cyan': 36,
        'white': 37,
        'default': 39,
        'on_black': 40,
        'on_red': 41,
        'on_green': 42,
        'on_yellow': 43,
        'on_blue': 44,
        'on_magenta': 45,
        'on_cyan': 46,
        'on_white': 47,
    }

    @classmethod
    def get_color(cls, text: str, color='default', style='default'):
        """
        Returns the text with the given style

        :param text: Text that will be written
        :type text: str
        :param color: Color of text, defaults to 'default'
        :type color: str, optional
        :param style: Style of text, defaults to 'default'
        :type style: str, optional
        :return: Text with styles and colors applied
        :rtype: str
        """
        return f'\x1b[{cls.styles[style]};{cls.colors[color]}m{text}\x1b[0m'



import regex
import re
import pickle
import spacy
import logging
import subprocess
import sys
import unicodedata
import nltk

from nltk.util import ngrams
from nltk.tokenize import word_tokenize

from pyspark.ml.feature import Tokenizer, StopWordsRemover, NGram, CountVectorizer, IDF
from pyspark.ml.pipeline import Pipeline
from pyspark.ml.linalg import SparseVector
from pyspark.sql import Window, DataFrame, SparkSession, Column, functions as F, types as T
from pyspark import StorageLevel
from api_model.utils.logger import logger


NLTK_STOPWORDS = nltk.corpus.stopwords.words('portuguese')
NLP = spacy.load("pt_core_news_sm")

MAX_COLS = {
    'negativity'
}
AVG_COLS = {'response_time'}

class NLExtractor: 

    def __init__(self):
      self.data_type =[]


    @classmethod
    def cleaner(self, text):
        
        #TO-DO: Keep Acronyms

        text = text.lower()
        text = re.sub(r",", " , ", text)
        text = re.sub(r"\'", " \' ", text)
        text = re.sub(r"!", " ! ", text)
        text = re.sub(r"\(", " ( ", text)
        text = re.sub(r"\)", " ) ", text)
        text = re.sub(r"\?", " ? ", text)
        text = re.sub(r"\:", " : ", text)
        text = re.sub(r"\.", " . ", text)
        text = re.sub(r"\$", "  ", text)
        text = re.sub(r"\-", "  ", text)
        text = re.sub(r"\s{2,}", " ", text)

        text = text.strip()

        return text


    @classmethod
    def remove_special_characters(self,text):
        try:
            return ''.join(
                c for c in unicodedata.normalize('NFD', text)
                if unicodedata.category(c) not in ['Mn', 'N'] and c.isalpha()
            ).lower()
        except IOError as e:
            print('Error tokenize', e)


    @classmethod
    def udf_split_text(self, text):
    
        if type(text) == str:
            text = text.encode()
        else :
            text = text.astype(str).encode()
        for x in text.decode().replace('\n', ' ').split(' '):
            yield x


    @classmethod
    def udf_clean_text(self,text):
        try:
            text = self.cleaner(text)
            logging.info(f'Sucess separed pontuation {text}')   
            out = []
            for x in self.udf_split_text(text):
                this_text = self.remove_special_characters(x)
                if len(this_text) > 2:
                    out.append(this_text)
            logging.info('Sucess clean_text')                    
            return ' '.join(' '.join(out).strip().split())
        except IOError:
            logging.info('Error clean_text')


    @classmethod
    def clean_text(self,text):
        try:
            text = self.cleaner(text)
            logging.info(f'Sucess separed pontuation {text}')   
            out = []
            for x in self.udf_split_text(text):
                if len(x) > 2:
                    out.append(x)
            logging.info('Sucess clean_text')                    
            return ' '.join(' '.join(out).strip().split())
        except IOError:
            logging.info('Error clean_text')


    @classmethod
    def tokenizer(self, text):
        #text = re.split(r'\s+',text)
        text = word_tokenize(text)
        return text


    @classmethod
    def filter_stop_words(self, text_tokens, additional_stop_words = []):
        stop_words = NLTK_STOPWORDS
        if len(additional_stop_words) > 0:
            for i in additional_stop_words:
                stop_words.append(i)
        #return [word for word in text_tokens if not word in stop_words]
        return ' '.join([word for word in text_tokens.split() if word not in stop_words])


    @classmethod
    def new_stopwords(self, text_tokens, additional_stop_words = []):
        stop_words = NLTK_STOPWORDS
        if len(additional_stop_words) > 0:
            for i in additional_stop_words:
                stop_words.append(i)        
        pattern = re.compile(r'\b(' + r'|'.join(stop_words) + r')\b\s*')
        text_tokens = pattern.sub('', text_tokens)
        return text_tokens

    
    @classmethod
    def convert_list_string(self,lista):
        out = []
        for i in lista:
            if len(i) > 2:
                out.append(i)
        return ' '.join(' '.join(out).strip().split())

    @classmethod
    def lemmatizer(self, lista, parser='spacy'):
        try:
            out = []
            lista = self.tokenizer(lista)
            for x in NLP(str(lista)):
                if ( not x.is_punct and
                    not x.is_space and
                    not x.is_stop and
                    len(x) > 1 and
                    self.remove_special_characters(x.text) != '' and 
                    self.remove_special_characters(x.text) != None ):
                    if x.pos_ == 'VERB':
                        out.append(self.remove_special_characters(x.lemma_))
                    else:
                        out.append(self.remove_special_characters(x.text))
            return ' '.join(' '.join(out).strip().split())
            #return list(dict.fromkeys(out))
        except IOError as e:
            print('Error text_lema', e)        
            pass


    def n_grams(self, lst, n_gram_size, filter_stop_words=False):
        n_grams_list = list(ngrams(lst,n_gram_size))
        return n_grams_list


    @classmethod
    def pattern_matcher(self, text, pattern, mode="dictionary", anonymizer = False, custom_anonymizer = '<UNK>'):

        output = []
        
        if mode == "regex":

            if isinstance(pattern, str):
                pattern = [pattern]

            for reg in pattern:
                if anonymizer:
                    text = re.sub(reg, custom_anonymizer, text)
                    output = text
                else:    
                    for match in re.finditer(reg, text):
                        output.append(self.detect_pattern(match.group().strip() , match.group(),match.group(), match.start(), match.end()))
        
        elif mode == "dictionary" or isinstance(pattern, dict):
            
            if isinstance(pattern, list):
                pattern = {p:[] for p in pattern}

            for key, synonims in pattern.items():

                pattern = r"(\b(" + "|".join([key] + synonims) + r")\b)"

                if anonymizer:
                    text = re.sub(pattern, custom_anonymizer, text)
                    output = text
                else:
                    if not re.search(r"\d", pattern):
                        pattern += r"{e<=1}"

                    for match in regex.finditer(pattern, text):
                        output.append(match.group().strip())
                        break
        else:
            print('Invalid Type.')
        
        output = self.convert_list_string(output)
        
        return output


    def ner(self,text, parser='spacy', ner_type = '', anonymizer = False, custom_anonymizer = '<UNK>'):
        
        output = []
        
        spacy_nlp = NLExtractor._instance.spacy_nlp
        
        if isinstance(ner_type, str):
            ner_type = [ner_type]
            
        if parser == 'spacy':

            if self._spacy_load == None:
                spacy_nlp = self.load_spacy_modules()
            else:
                spacy_nlp = self._spacy_load

            if spacy_nlp == None:
                return []

            ner = spacy_nlp(text)
            for entidade in ner.ents:   

                if anonymizer:
                    if len(ner_type) == 0:
                        print('WARNING: Please Specify An Entity Type.')
                        return text

                    for ner in ner_type:
                        if entidade.label_ == ner:
                            try:
                                text = re.sub(entidade.text, custom_anonymizer, text)
                            except:
                                text = text
                    output = text
                else:
                    output.append(
                        self.detect_pattern(entidade.text, entidade.text, entidade.label_, entidade.start_char, entidade.end_char)
                    )
        else:
            print('Invalid Parser.')

        return output


    @classmethod
    def detect_pattern(self, match, key, ner_type, start_index, end_index):
        # TODO: add score 
        return {
            "entity": match,
            "value": key,
            "type": ner_type, 
            "startIndex": start_index,
            "endIndex": end_index,
        }       


    def load_spacy_modules(self):
        try:
            self._spacy_load = spacy.load("pt_core_news_sm")
        except: 
            #TO-DO: Recognize and install spacy packages on runtime
            print('Warning!!!')
            print('Please Run the following command:')
            print('python -m spacy download pt_core_news_sm')
            print('before using spaCy modules')
            _spacy_load = None


    @classmethod
    def udf_type_keywords(self, text, pattern,  mode="dictionary"):
        """
        Search pattern in text and return what was found separated by pipe
        """

        if mode == "dictionary" or isinstance(pattern, dict):
            if isinstance(pattern, list):
                pattern = {p:[] for p in pattern}
                mentions = []
                for word in pattern:
                    regx = fr'\b{word}\b'

                    n_mentions = len(re.findall(regx, text))
                    if n_mentions:
                        mentions.append(word)
                return '|'.join(mentions)
        return None


    @classmethod
    def udf_extract_digits(self, text, pattern):
        document_terms = []
        
        cont_pattern = pattern
        for match in re.finditer(cont_pattern, str(text),re.IGNORECASE):
            document_terms.append(match.group().strip())

        document_terms = self.remove_duplicates(document_terms)
        document_terms = [re.sub('[^a-z0-9 ]+', '', str(document_terms))]
        return ''.join(','.join(document_terms).strip()).split()


    @classmethod
    def remove_duplicates(self, lista):
        l = []
        for i in lista:
            if i not in l:
                l.append(i)
        l.sort()
        return l


    def remove_specific_numbers(self, lista, numbers):
        return [token for token in lista if token not in numbers]

 
    @F.udf(T.StringType())
    def most_important_ngram(vocabulary, v):
        """
        Spark UDF that extracts the most relevant words, 2-grams and 3-grams with TF-IDF

        :param Vocabulary: Vocabulary of words by CountVectorizerModel
        :type vocabulary: str
        :param v: SparseVector with TF-IDF relevance values
        :type v: SparceVector
        :return: Dictionary with most relevant word, 2-gram anda 3-gram
        :rtype: dict
        """

        if isinstance(v, SparseVector):
            kv = {vocabulary[i]: cnt for (i, cnt) in zip(v.indices.tolist(), v.values.tolist())}
            return max(kv, key=kv.get, default='')
        return ''


    def spark(mode='default') -> SparkSession:
        """Retrieve current spark session.

        :param mode: str
            The session mode. Should be either "default" or "test".

        :return: SparkSession
        """
        logger.info(f'[INFO] Creating {mode} Spark Session')
        if mode == 'default':
            return SparkSession.builder.config("spark.driver.memory", "12g").getOrCreate()
        else:
            raise ValueError(f'Illegal value "{mode}" mode parameter. '
                            'It should be either "default", "test" or "deltalake".')        

    @classmethod
    def most_relevant_ngram(self,
        x: DataFrame,
        text_column: str,
        output_column_prefix: str,
        id_field: str = 'id',
        where: Column = None,
        stop_words='portuguese',
        features: int = 4096,
        min_df: float = 3.0,
        keep_intermediate: bool = False,
        texts_to_filter: str = (),
    ):
        """
        Computes and creates features to select the most relevant word, 2-grams and 3-grams.

        :param x: DataFrame Input containing pre-processed text
        :type x: DataFrame
        :param text_column: Name of the column containing the evaluating documents
        :type text_column: str
        :param output_column_prefix: Prefix used to create the output column.
        :type output_column_prefix: str
        :param id_field: Name of the column of the unique id to join the result on the original data frame in
                        case filter_expr was not null, defaults to 'id'
        :type id_field: str, optional
        :param where: Expression used to filter the dataframe if you want to extract the most relevant ngrams for a specific context, defaults to None
        :type where: column, optional
        :param stop_words: List of words fed to `StopWordsRemover` transformer, defaults to 'portuguese'
        :type stop_words: str, optional
        :param features: Max size of the vocabulary, defaults to 4096
        :type features: int, optional
        :param min_df: Specifies the minimum number of different documents a term must appear in to be included in the vocabulary
        :type min_df: float, optional
        :param keep_intermediate: Retain intermediate representation
        :type keep_intermediate: bool, optional
        :return: DataFrame containing most relevant word, 2-grams and 3-grams columns
        :rtype: DataFrame
        """

        o = x

        if isinstance(stop_words, str):
            stop_words = StopWordsRemover.loadDefaultStopWords('portuguese')

        if where is not None:
            x = o.filter(where)

        tokenizer = Tokenizer(inputCol=text_column, outputCol=f'{text_column}_aux_tokenized')

        words_remover = StopWordsRemover(
            inputCol=tokenizer.getOutputCol(), outputCol=f'{text_column}_aux_filtered', stopWords=stop_words
        )

        bigram = NGram(n=2, inputCol=words_remover.getOutputCol(), outputCol=f'{text_column}_aux_bigrams')
        trigram = NGram(n=3, inputCol=words_remover.getOutputCol(), outputCol=f'{text_column}_aux_trigrams')
        words_cv = CountVectorizer(
            inputCol=words_remover.getOutputCol(), outputCol=f'{text_column}_aux_words_cv', vocabSize=features, minDF=min_df
        )

        bigram_cv = CountVectorizer(
            inputCol=bigram.getOutputCol(), outputCol=f'{text_column}_aux_bigrams_cv', vocabSize=features, minDF=min_df
        )

        trigram_cv = CountVectorizer(
            inputCol=trigram.getOutputCol(), outputCol=f'{text_column}_aux_trigrams_cv', vocabSize=features, minDF=min_df
        )

        words_idf = IDF(inputCol=words_cv.getOutputCol(), outputCol=f'{output_column_prefix}_words_idf')
        bigram_idf = IDF(inputCol=bigram_cv.getOutputCol(), outputCol=f'{output_column_prefix}_bigrams_idf')
        trigram_idf = IDF(inputCol=trigram_cv.getOutputCol(), outputCol=f'{output_column_prefix}_trigrams_idf')

        pipeline = Pipeline(
            stages=[
                tokenizer,
                words_remover,
                bigram,
                trigram,
                words_cv,
                bigram_cv,
                trigram_cv,
                words_idf,
                bigram_idf,
                trigram_idf,
            ]
        )

        model = pipeline.fit(x)
        x = model.transform(x)

        word_field = f'{output_column_prefix}_word'
        bi_gram_field = f'{output_column_prefix}_bigram'
        tri_gram_field = f'{output_column_prefix}_trigram'

        for field, col, vocabulary in (
            (word_field, words_idf.getOutputCol(), model.stages[4].vocabulary),
            (bi_gram_field, bigram_idf.getOutputCol(), model.stages[5].vocabulary),
            (tri_gram_field, trigram_idf.getOutputCol(), model.stages[6].vocabulary),
        ):
            x = x.withColumn(field, self.most_important_ngram(F.array([F.lit(x) for x in vocabulary]), col))

        if not keep_intermediate:
            x = x.drop(
                bigram.getOutputCol(),
                trigram.getOutputCol(),
                tokenizer.getOutputCol(),
                words_remover.getOutputCol(),
                bigram_cv.getOutputCol(),
                trigram_cv.getOutputCol(),
                words_cv.getOutputCol(),
                words_idf.getOutputCol(),
                bigram_idf.getOutputCol(),
                trigram_idf.getOutputCol(),
            )

        x.persist(StorageLevel.DISK_ONLY)
        x.count()

        o.persist(StorageLevel.DISK_ONLY)
        o.count()

        if where is not None:
            id_r = f'filtered_{id_field}'
            x = x.select(id_field, bi_gram_field, tri_gram_field, word_field).withColumnRenamed(id_field, id_r)

            x = o.join(x, o[id_field] == x[id_r], how='left').drop(id_r)

        return x, model
    #TO-DO: Datetime Converter
    
    @classmethod
    def group_df(self, df, interlocutor, message_content):
        print('Grouping messages into one unique record')

        schema = []

        logger.debug(f'Message Author Column: message_author')
        logger.debug(f'Message content Column: {message_content}')

        bot_identifiers = interlocutor.get('bot_identifiers', [])
        client_identifiers = interlocutor.get('client_identifiers', [])

        logger.debug(f'{bot_identifiers}, {client_identifiers}')

        has_attendant = F.when(
            F.col('message_author').isin(bot_identifiers + client_identifiers), F.lit(0)
        ).otherwise(F.lit(1))

        df = df.withColumn('has_attendant', has_attendant)

        max_columns = [col for col in df.columns if col.endswith('_findint')]

        logger.debug(f'separadores cliente e operador, {interlocutor}')

        msgs_df = self._group_messages_df(df, bot_identifiers, interlocutor)
        kpis_df = self._group_kpis_df(df, max_columns, 'avg_columns')


        return kpis_df.join(msgs_df, on='issue_id', how='full')


    @classmethod
    def _group_messages_df(self, df, bot_identifiers, separators):
        all_message_columns = self._get_all_messages_columns(df, bot_identifiers, separators)

        return (
            df.orderBy('issue_id', 'message_order')
            .groupBy('issue_id')
            .agg(
                *all_message_columns,
                F.array_join(
                    F.collect_list(
                        F.concat(
                            F.col('message_time'),
                            F.lit(' - Author: '),
                            F.col('message_author'),
                            F.lit(' - Message: '),
                            F.col('original_message'),
                        )
                    ),
                    '\n---\n',
                ).alias('original_messages'),
            )
        )

    @classmethod
    def _group_kpis_df(self, df, additional_max_cols: list, additional_avg_cols: list):
        ignore_columns = ['issue_id', 'message_content', 'original_message']

        group_max_columns = MAX_COLS.union(additional_max_cols).intersection(df.columns)
        group_avg_columns = AVG_COLS.union(additional_avg_cols).intersection(df.columns)
        group_columns = set(df.columns).difference(group_max_columns.union(group_avg_columns).union(ignore_columns))

        logger.debug(f'MAX columns: {group_max_columns}')
        logger.debug(f'AVG columns: {group_avg_columns}')
        logger.debug(f'FIRST columns: {group_columns}')

        # !TODO: first is non-deterministic, could be getting something other than first record
        kpis_df = (
            df.orderBy('issue_id', 'message_order')
            .groupBy('issue_id')
            .agg(
                *(F.first(col).alias(col) for col in group_columns),
                *(F.max(col).alias(col) for col in group_max_columns),
                *(F.round(F.avg(col)).alias(f'average_{col}') for col in group_avg_columns),
            )
        )

        return kpis_df

    @classmethod
    def _get_all_messages_columns(self, df, bot_identifiers, separators):
        all_messages_col = self._join_messages(df, 'all_messages', bot_identifiers)

        columns = [all_messages_col]

        for col_separator, vals_separator in separators.items():
            for val_separator in vals_separator:
                first_value = val_separator.split('|')[0]
                alias = f'all_messages_{col_separator}_{first_value}'
                condition = F.col(col_separator).isin(val_separator.split('|'))

                current_all_messages_col = self._join_messages(df, alias, bot_identifiers, condition)

                columns.append(current_all_messages_col)

        return columns

    @classmethod
    def _join_messages(self, df, alias, bot_identifiers, condition=None):
        if condition is None:
            condition = F.lit(True)

        logger.debug(f'Joining messages to {alias}')      

        return F.array_join(
            F.collect_list(
                F.when(
                    (~F.col('message_author').isin(bot_identifiers)) & (condition),
                    F.col('message_content'),
                )
            ),
            ' ',
        ).alias(alias)


    @classmethod
    def populate_columns(self,df, columns):
        """
        Creates and fills the specified columns with values that are deducted from other columns that are present and correctly filled in the DataFrame

        :param df: Input DataFrame
        :type df: DataFrame
        :param columns: Columns that will be created and filled
        :type columns: list
        :return: DataFrame with the specified columns created and filled
        :rtype: DataFrame
        """

        if 'message_order' in columns:
            df = self.populate_message_order(df)
        return df

    @classmethod
    def populate_message_order(self, df, id_field = 'issue_id', message_time = 'message_time'):
        """
        Creates and fills the `message_order` column

        Based on the id of the conversation and the message time, the order of the messages is deducted and filled

        :param df: Input DataFrame
        :type df: DataFrame
        :param id_field: The column that has the conversation id, defaults to 'issue_id'
        :type id_field: str
        :param message_time: The column that contains the time of each message, defaults to 'message_time'
        :type message_time: str
        :return: DataFrame with the message_order filled
        :rtype: DataFrame
        """

        print('Generating `message_order` column')
        win = Window.partitionBy(id_field).orderBy(F.col(message_time).asc())
        df = df.withColumn('message_order', F.row_number().over(win))
        return df



from datetime import datetime
from xmlrpc.client import DateTime
import pandas as pd
from api_model.nlextract import NLExtractor
from pyspark.sql import functions as F, types as T
from api_model.utils.logger import logger
from datetime import date
import os
import re
from pytz import timezone


sao_paulo_timezone = timezone('America/Sao_Paulo')


PATH_READ = '/content/'
PATH_SAVE = '/content/drive/My Drive/'

#PATH_READ = '/opt/dna/find-keywords/datalake/'
#PATH_SAVE = '/opt/dna/find-keywords/datalake/'


class TransforDatas(NLExtractor):
    def __init__(self):
        self.data_type =[]
    
    @classmethod
    def convert_dataframe(self, df, id_database, column_text, response_time, filename, prefix, prefix_sep, interlocutor, encoding):

        column_name = ' '.join([str(item) for item in list(interlocutor.keys())])
        logger.info(f'rename valus of interlocutor column: {column_name}')

        logger.info('convert dataframe pandas to pyspark')
        df = df.rename(columns={id_database:'issue_id', column_text:'message_content', response_time:'message_time', column_name:'message_author'})

        logger.debug(f'save temp file')
        df = self.save_file(df=df, filename=filename, meth='temp')

        logger.debug(f'load temp file')
        dfPyspark = self.load_file(prefix=prefix, filename=filename, prefix_sep=prefix_sep, encoding=encoding, meth='temp')

        logger.info('created message order')
        dfPyspark = self.populate_message_order(dfPyspark, id_field='issue_id', message_time='message_time')    

        logger.debug(f'print new columns of pyspark dataframe: {dfPyspark.printSchema()}')

        return dfPyspark


    @classmethod
    def filter_columns(self, df):

        out = ['issue_id','message_content','message_time', 'message_author']
        list_1 = [col for col in df.columns if col in out]
        list_2 = [col for col in df.columns if col.endswith('_findint')]
        columns_to_extends = []

        for i in list_1:
            columns_to_extends.append(i)
        
        for y in list_2:
            columns_to_extends.append(y)

        logger.debug(f'get only columns to findkey words {columns_to_extends}')

        return columns_to_extends

    @classmethod
    def statistics_dataframe(self, df, column_text):

        logger.info('check numbers words by rows')
        df['numbers_words'] = df[column_text].apply(lambda x: len(str(x).split(' ')))

        df_mim = df[df['numbers_words']<=3]
        logger.info(f'numbers of rows < 3 words from line {df_mim["numbers_words"].count()}')

        df_max = df[df['numbers_words']>3]
        logger.info(f'numbers of promotors lines {df_max["numbers_words"].count()}')

        return df

    @classmethod
    def text_mining(self, df, column_text):

        logger.info(f'remove special characters and pontuation of column_text')
        df[column_text] =  df[column_text].apply(lambda x: self.udf_clean_text(x))

        logger.info('tranform text in text lemma')
        df[column_text] =  df[column_text].apply(lambda x: self.lemmatizer(x))

        return df

    @classmethod
    def stop_words_text(self, df, column_text, additional_stop_words):

        logger.info('remove stop words from text')
        logger.info(f'result before of process stop words \n {df[column_text].head(5)}')
        df[column_text] =  df[column_text].apply(lambda x: self.filter_stop_words(x, additional_stop_words))
        logger.info(f'result after of process stop words \n {df[column_text].head(5)}')

        return df

    @classmethod
    def word_search(self, df, list_pattern, type_find, column_text):

        logger.info('collect words and find in column_text')
        logger.debug(f'dict: {list_pattern}')
        try:
            for key in list_pattern:
                if type_find == 'fixo':
                    df[f'{key}_findint'] =  df[column_text].apply(lambda x: self.udf_type_keywords(x,list_pattern[key],mode="dictionary"))
                else:
                    df[f'{key}_findint'] =  df[column_text].apply(lambda x: self.pattern_matcher(x,list_pattern[key],mode="dictionary"))
        except IOError as e:
            logger.debug(f'não tem mais listas para rodar dados {e}')
        pass

        return df

    @classmethod
    def merge_dataframe(self, df, df_all):

        logger.info('convert to pandas again')
        df = df.toPandas()

        logger.debug(f'dataframe old {list(df_all.columns)}')
        logger.debug(f'dataframe new {list(df.columns)}')      
        
        df_merge = pd.concat([df_all, df], axis=1)
        logger.debug(f'joined dataframes {list(df_merge.columns)}  \n and numbers of rows is {df_merge["numbers_words"].count()}')

        return df_merge

    @classmethod
    def normalize_data(self, df, column_text, response_time, format_data, id_database):

        logger.info('convert column_text column to string type')
        df[column_text] = df[column_text].astype(str)    

        logger.info('put column_text in lower case')
        df[column_text] = df[column_text].str.lower()

        logger.info('normalize id column')
        df[id_database] = df[id_database].astype(str)   

        if response_time != '':
            logger.info('convert date column string in date column timestamp')
            df[response_time] = df[response_time].apply(lambda x: self.convert_datetime(x, format_data))
            logger.debug(f'dataformat: {df[response_time].head(5)}')
        else:
            logger.info('create date column timestamp')
            df[response_time] =pd.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            logger.debug(f'dataformat: {df[response_time].head(5)}')

        return df


    @classmethod
    def convert_datetime(self, dateformat, formats):
        
        if dateformat and isinstance(dateformat, str):
            for dtformat in formats.split('|'):
                try:
                    time = datetime.astimezone(sao_paulo_timezone).strftime(dateformat, dtformat)
                    return time
                except Exception:
                    continue
        elif isinstance(dateformat, datetime):
            return dateformat

        return None

    
    @classmethod
    def save_file(self, df, filename, meth):
        
        if meth == 'temp':

            logger.info('save csf file')
            filename_renomead = f'{filename}_temp'
            file_save = f'{filename_renomead}.csv'
            df.to_csv(f'{PATH_SAVE}/{file_save}', sep=';',encoding='utf-8',index=False)
            logger.info('Finishing Process')

        else:

            logger.info('save csf file')
            filename_renomead = f'{filename}_tratado'
            file_save = f'{filename_renomead}.csv'
            df.to_csv(f'{PATH_SAVE}/{file_save}', sep=';',encoding='utf-8',index=False)
            logger.info('Finishing Process') 

        return df

    
    @classmethod
    def load_file(self, prefix, filename, prefix_sep, encoding, meth):

        if meth == 'temp':

            logger.info('load csf file temp')
            filename_renomead = f'{filename}_temp'
            file_save = f'{filename_renomead}.csv'
            spark = self.spark('default')

            df = (spark.read.option('quote', '\"')
                    .option('escape', '\"')
                    .option('multiline', 'true')
                    .option('header', 'true')
                    .option('sep', ';')
                    .csv(f'{PATH_SAVE}/{file_save}'))

            logger.info('Finishing Process')

            return df        

        else:

            logger.info(f'read file {PATH_READ}{prefix}')
            if prefix == 'xlsx':
                df = pd.read_excel(f"{PATH_READ}/{filename}.{prefix}", engine='openpyxl')
                logger.debug(f'Schema of dataframe is {df.info()}')
            if prefix == 'csv':
                df = pd.read_csv(f"{PATH_READ}/{filename}.{prefix}", sep=prefix_sep, encoding=encoding)
                logger.debug(f'Schema of dataframe is {df.info()}')

        return df

    
    @classmethod
    def delete_file(self, filename):

        filepath = f'{PATH_SAVE}/{filename}.csv'

        if os.path.exists(filepath):
            logger.info(f'deleting file: {filepath}')
            os.remove(filepath)
        
        else:
            logger.info("Can not delete the file as it doesn't exists")

        return None
     



from setuptools import find_namespace_packages, setup
from setuptools.command.install import install
from setuptools import setup, find_packages
from distutils.command.install import install as _install

import os


PROJECT_DIR = os.path.dirname(__file__)
INFO = open(os.path.join(PROJECT_DIR, 'INFO')).readlines()
INFO = dict((line.strip().split('=') for line in INFO))

DEPENDENCIES = open(os.path.join(PROJECT_DIR, 'requirements.txt')).readlines()

setup(
    name='model_fkeywords',
    description='A Natural Language Processing Library',
    version=INFO['version'],
    author=INFO['author'],
    author_email=INFO['author_email'],
    url=INFO['url'],    
    license=open(os.path.join(PROJECT_DIR, 'LICENSE')).read(),
    packages=find_namespace_packages(include=['api_model','api_model.utils']),
    install_requires=[d for d in DEPENDENCIES if '://' not in d],
    python_requires='==3.7.13',
    #TO-DO: Fix dependency links : not working with bdist_wheel
    dependency_links = ["git+https://github.com/explosion/spacy-models/releases/download/pt_core_news_sm-3.2.0/pt_core_news_sm-3.2.0.tar.gz"],
    tests_require=['pytest', 'parameterized'],
    zip_safe=False
)

